@@ -8,9 +8,13 @@
 """
 Main Qt4 application for hgview
 """
+
 import sys, os
+import os.path as osp
 import re
 import errno
+from functools import partial
+from operator import methodcaller
 
 from PyQt4 import QtCore, QtGui, Qsci
 
@@ -20,10 +24,12 @@ from mercurial.error import RepoError
 
 
 from hgviewlib.application import HgRepoViewer as _HgRepoViewer
-from hgviewlib.util import tounicode, find_repository, rootpath, build_repo
+from hgviewlib.util import (tounicode, find_repository, rootpath, build_repo,
+                            upward_path, read_nested_repo_paths, allbranches)
 from hgviewlib.hggraph import diff as revdiff
 from hgviewlib.decorators import timeit
 from hgviewlib.config import HgConfig
+from hgviewlib.util import compose
 
 from hgviewlib.qt4 import icon as geticon
 from hgviewlib.qt4.hgrepomodel import HgRepoListModel, HgFileListModel
@@ -38,9 +44,8 @@ from mercurial.error import RepoError
 
 Qt = QtCore.Qt
 bold = QtGui.QFont.Bold
-connect = QtCore.QObject.connect
-SIGNAL = QtCore.SIGNAL
 
+NESTED_PREFIX = u'\N{RIGHTWARDS ARROW} '
 
 class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
     """hg repository viewer/browser application"""
@@ -55,9 +60,9 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         QtGui.QApplication.setStyle(QtGui.QStyleFactory.create('Cleanlooks'))
         QtGui.QMainWindow.__init__(self)
         if not self.repo.root:
-            repo = self.ask_repository()
-            if repo:
-                self.repo = repo
+            repopath = self.ask_repository()
+            if repopath:
+                self.repo = build_repo(ui.ui(), repopath)
         HgDialogMixin.__init__(self)
 
         self.setWindowTitle('hgview: %s' % os.path.abspath(str(self.repo.root)))
@@ -73,9 +78,9 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.createToolbars()
 
         self.textview_status.setFont(self._font)
-        connect(self.textview_status, SIGNAL('showMessage'),
+        self.textview_status.message_logged.connect(
                 self.statusBar().showMessage)
-        connect(self.tableView_revisions, SIGNAL('showMessage'),
+        self.tableView_revisions.message_logged.connect(
                 self.statusBar().showMessage)
 
         # setup tables and views
@@ -84,15 +89,17 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
             self.setupHeaderTextview()
             self.setupBranchCombo()
             self.setupModels(fromhead)
- 
-        if self.cfg.getFileDescriptionView() == 'asfile':
-            fileselcallback = self.displaySelectedFile
-        else:
-            fileselcallback = self.textview_status.displayFile
-        connect(self.tableView_filelist, SIGNAL('fileSelected'),
-                fileselcallback)
+            self._setupQuickOpen()
 
-        connect(self.textview_status, SIGNAL('revForDiffChanged'),
+        to_utf8 = methodcaller('encode', 'utf-8')
+        if self.cfg.getFileDescriptionView() == 'asfile':
+            fileselcallback = compose(self.displaySelectedFile, to_utf8)
+        else:
+            fileselcallback = compose(self.textview_status.displayFile, to_utf8)
+        self.tableView_filelist.file_selected[str].connect(fileselcallback)
+        self.tableView_filelist.file_selected[str, int].connect(fileselcallback)
+
+        self.textview_status.rev_for_diff_changed.connect(
                 self.textview_header.setDiffRevision)
 
         if fromhead:
@@ -112,20 +119,52 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
 
                 self.reload()
 
+    def _setupQuickOpen(self):
+        """Initiliaze the quick open menu
+
+        This call utils function to search for possible nested repo setup"""
+        # quick open repository
+        self.quickOpen_comboBox.clear()
+        repopath = osp.abspath(self.repo.root)
+        #  search backward until the root folder for a master
+        #  repository that contains confman or subrepo data
+        for master_path in upward_path(repopath):
+            if not osp.exists(osp.join(master_path, '.hg')):
+                continue  # Not a repo!
+            subrepos = read_nested_repo_paths(master_path)
+            if not subrepos:
+                continue  # Nothing nested!
+            # We found a nested repo setup!
+            #
+            # But is it related to the viewed repo?
+            involved_paths = [master_path] + [pth for _, pth in subrepos]
+            if repopath in involved_paths:
+                master_name = tounicode(osp.basename(master_path))
+                repos = [(master_name, master_path)]
+                for name, path in subrepos:
+                    repos.append((NESTED_PREFIX + name, path))
+                repos.sort()
+                break  # I've found related! I'll survive the glaciation!
+        else:
+            # They migrated without me. They do this every year.
+            #
+            # Nothing to quick-open hide the toolbar and interrupt
+            self.toolBar_quickopen.setVisible(False)
+            return
+        # TODO: add a treeview with a complete graph.
+        curidx = 0
+        for idx, (text, data) in enumerate(repos):
+            if data == self.repo.root:
+                curidx = idx
+            self.quickOpen_comboBox.addItem(text, data)
+        self.quickOpen_comboBox.setCurrentIndex(curidx)
+
     def setupBranchComboAndReload(self, *args):
         self.setupBranchCombo()
         self.reload()
 
     def setupBranchCombo(self, *args):
-        allbranches = sorted(self.repo.branchtags().items())
-
-        openbr = []
-        for branch, brnode in allbranches:
-            openbr.extend(self.repo.branchheads(branch, closed=False))
-        clbranches = [br for br, node in allbranches if node not in openbr]
-        branches = [br for br, node in allbranches if node in openbr]
-        if self.branch_checkBox_action.isChecked():
-            branches = branches + clbranches
+        branches = allbranches(self.repo, self.branch_checkBox_action.isChecked())
 
         if len(branches) == 1:
             self.branch_label_action.setEnabled(False)
@@ -141,11 +180,13 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.find_toolbar = FindInGraphlogQuickBar(self)
         self.find_toolbar.attachFileView(self.textview_status)
         self.find_toolbar.attachHeaderView(self.textview_header)
-        connect(self.find_toolbar, SIGNAL('revisionSelected'),
+        self.find_toolbar.revision_selected.connect(
                 self.tableView_revisions.goto)
-        connect(self.find_toolbar, SIGNAL('fileSelected'),
+        self.find_toolbar.revision_selected[int].connect(
+                self.tableView_revisions.goto)
+        self.find_toolbar.file_selected.connect(
                 self.tableView_filelist.selectFile)
-        connect(self.find_toolbar, SIGNAL('showMessage'),
+        self.find_toolbar.message_logged.connect(
                 self.statusBar().showMessage,
                 Qt.QueuedConnection)
 
@@ -160,6 +201,12 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         findaction.setIcon(geticon('find'))
         self.toolBar_edit.addAction(findaction)
 
+        cb = self.quickOpen_comboBox = QtGui.QComboBox()
+        cb.setStatusTip("Quick open other repositories")
+        self.quickOpen_action = self.toolBar_quickopen.addWidget(cb)
+        _callback = lambda: self.open_repository(cb.itemData(cb.currentIndex()))
+        self.quickOpen_comboBox.activated.connect(_callback)
+
         # tree filters toolbar
         self.toolBar_treefilters.addAction(self.actionShowHidden)
         self.toolBar_treefilters.addAction(self.tableView_revisions._actions['unfilter'])
@@ -175,10 +222,8 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.branch_checkBox_action = cbranch_action
         self.branch_label.setMenu(self.branch_menu)
         self.branch_comboBox = QtGui.QComboBox()
-        connect(self.branch_comboBox, SIGNAL('activated(const QString &)'),
-                self.refreshRevisionTable)
-        connect(cbranch_action, SIGNAL('toggled(bool)'),
-                self.setupBranchComboAndReload)
+        self.branch_comboBox.activated[str].connect(self.refreshRevisionTable)
+        cbranch_action.toggled[bool].connect(self.setupBranchComboAndReload)
 
         self.toolBar_treefilters.layout().setSpacing(3)
 
@@ -199,11 +244,10 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         follow_action.setStatusTip("Follow changeset history from start revision")
         self.startrev_follow_action = follow_action
         self.startrev_label.setMenu(self.startrev_menu)
-        callback = lambda *a: self.tableView_revisions.emit(
-                    SIGNAL('startFromRev'),
-                    self.startrev_entry, self.startrev_follow_action)
-        connect(self.startrev_entry, SIGNAL('editingFinished()'), callback)
-        connect(self.startrev_follow_action, SIGNAL('toggled(bool)'), callback)
+        callback = lambda *a: self.tableView_revisions.start_from_rev[str, bool].emit(
+            str(self.startrev_entry.text()), self.startrev_follow_action.isChecked())
+        self.startrev_entry.editingFinished.connect(callback)
+        self.startrev_follow_action.toggled[bool].connect(callback)
 
         self.revscompl_model = QtGui.QStringListModel(['tip'])
         self.revcompleter = QtGui.QCompleter(self.revscompl_model, self)
@@ -213,33 +257,68 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.startrev_entry_action = self.toolBar_treefilters.addWidget(self.startrev_entry)
 
         # diff mode toolbar
-        actions = self.textview_status.sci._actions
+        actions = self.textview_status.sci.actions
         for action_name in ('diffmode', 'prev', 'next', 'show-big-file'):
             self.toolBar_diff.addAction(actions[action_name])
-        self.toolBar_diff.setVisible(self.cfg.getToolBarDiffAtStartup())
 
         # rev mod toolbar
         if self.textview_header.rst_action is not None:
             self.toolBar_rev.addAction(self.textview_header.rst_action)
-        self.toolBar_rev.setVisible(self.cfg.getToolBarRevAtStartup())
+
+        self._handle_toolbar_visibility()
+
+        self.addAction(self.actionQuit)
+
+    def _handle_toolbar_visibility(self):
+        """Initial value and event hooking
+
+        This function read toolbar related persistent settings from QT API.
+        It also setup hooks on visibility changes so to make the setting
+        persistent.
+        """
+        toolbars = (self.toolBar_file,
+                    self.toolBar_quickopen,
+                    self.toolBar_edit,
+                    self.toolBar_treefilters,
+                    self.toolBar_diff,
+                    self.toolBar_rev,
+                    self.toolBar_help)
+        settings = QtCore.QSettings()
+        for toolbar in toolbars:
+            entryname = '%s/%s/visible' % (self.objectName(), toolbar.objectName())
+            # bring back persistent status
+            status = settings.value(entryname)
+            if status is None:
+                status = 1
+            else:
+                try:
+                    status = int(status)
+                except ValueError:
+                    # for backward compatibility
+                    status = {'false': False , 'true': True}.get(status.lower(), status)
+                    status = int(status)
+
+            toolbar.setVisible(status)
+            # update settings on visibility toggleling
+            # We use integers because PyQt4 setting stores boolean as strings
+            # with QVariant api version 2
+            toolbar.toggleViewAction().toggled[bool].connect(
+                compose(partial(settings.setValue, entryname), int)
+            )
+
 
     def createActions(self):
         # main window actions (from .ui file)
-        connect(self.actionOpen_repository, SIGNAL('triggered()'),
-                self.open_repository)
-        connect(self.actionRefresh, SIGNAL('triggered()'),
-                self.reload)
-        connect(self.actionAbout, SIGNAL('triggered()'),
-                self.on_about)
-        connect(self.actionQuit, SIGNAL('triggered()'),
-                self.close)
+        self.actionOpen_repository.triggered.connect(self.open_repository)
+        self.actionRefresh.triggered.connect(self.reload)
+        self.actionAbout.triggered.connect(self.on_about)
+        self.actionQuit.triggered.connect(self.close)
         self.actionQuit.setIcon(geticon('quit'))
         self.actionRefresh.setIcon(geticon('reload'))
 
         self.actionHelp.setShortcut(Qt.Key_F1)
         self.actionHelp.setIcon(geticon('help'))
-        connect(self.actionHelp, SIGNAL('triggered()'),
-                self.on_help)
+        self.actionHelp.triggered.connect(self.on_help)
 
         self.actionShowHidden = QtGui.QAction(self.tr('Show/Hide Hidden'), self)
         self.actionShowHidden.setIcon(geticon('showhide'))
@@ -247,8 +326,7 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.actionShowHidden.setChecked(self.cfg.getShowHidden())
         self.actionShowHidden.setToolTip(self.tr('Show/Hide hidden changeset'))
         self.actionShowHidden.setStatusTip(self.tr('Show/hide hidden changeset'))
-        connect(self.actionShowHidden, SIGNAL('triggered()'),
-                self.refreshRevisionTable)
+        self.actionShowHidden.triggered.connect(self.refreshRevisionTable)
 
         self.actionShowMainContent = QtGui.QAction('Content', self)
         self.actionShowMainContent.setIcon(geticon('content'))
@@ -258,18 +336,15 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.actionShowMainContent.setToolTip(tip)
         self.actionShowMainContent.setStatusTip(tip)
         self.actionShowMainContent.setShortcut(Qt.Key_Space)
-        connect(self.actionShowMainContent, SIGNAL('triggered()'),
-                self.toggleMainContent)
+        self.actionShowMainContent.triggered.connect(self.toggleMainContent)
 
         # Next/Prev file
         self.actionNextFile = QtGui.QAction('Next file', self)
         self.actionNextFile.setShortcut('Right')
-        connect(self.actionNextFile, SIGNAL('triggered()'),
-                self.tableView_filelist.nextFile)
+        self.actionNextFile.triggered.connect(self.tableView_filelist.nextFile)
         self.actionPrevFile = QtGui.QAction('Prev file', self)
         self.actionPrevFile.setShortcut('Left')
-        connect(self.actionPrevFile, SIGNAL('triggered()'),
-                self.tableView_filelist.prevFile)
+        self.actionPrevFile.triggered.connect(self.tableView_filelist.prevFile)
         self.addAction(self.actionNextFile)
         self.addAction(self.actionPrevFile)
         self.disab_shortcuts.append(self.actionNextFile)
@@ -278,12 +353,10 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         # Next/Prev rev
         self.actionNextRev = QtGui.QAction('Next revision', self)
         self.actionNextRev.setShortcut('Down')
-        connect(self.actionNextRev, SIGNAL('triggered()'),
-                self.tableView_revisions.nextRev)
+        self.actionNextRev.triggered.connect(self.tableView_revisions.nextRev)
         self.actionPrevRev = QtGui.QAction('Prev revision', self)
         self.actionPrevRev.setShortcut('Up')
-        connect(self.actionPrevRev, SIGNAL('triggered()'),
-                self.tableView_revisions.prevRev)
+        self.actionPrevRev.triggered.connect(self.tableView_revisions.prevRev)
         self.addAction(self.actionNextRev)
         self.addAction(self.actionPrevRev)
         self.disab_shortcuts.append(self.actionNextRev)
@@ -292,23 +365,19 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         # navigate in file viewer
         self.actionNextLine = QtGui.QAction('Next line', self)
         self.actionNextLine.setShortcut(Qt.SHIFT + Qt.Key_Down)
-        connect(self.actionNextLine, SIGNAL('triggered()'),
-                self.textview_status.nextLine)
+        self.actionNextLine.triggered.connect(self.textview_status.nextLine)
         self.addAction(self.actionNextLine)
         self.actionPrevLine = QtGui.QAction('Prev line', self)
         self.actionPrevLine.setShortcut(Qt.SHIFT + Qt.Key_Up)
-        connect(self.actionPrevLine, SIGNAL('triggered()'),
-                self.textview_status.prevLine)
+        self.actionPrevLine.triggered.connect(self.textview_status.prevLine)
         self.addAction(self.actionPrevLine)
         self.actionNextCol = QtGui.QAction('Next column', self)
         self.actionNextCol.setShortcut(Qt.SHIFT + Qt.Key_Right)
-        connect(self.actionNextCol, SIGNAL('triggered()'),
-                self.textview_status.nextCol)
+        self.actionNextCol.triggered.connect(self.textview_status.nextCol)
         self.addAction(self.actionNextCol)
         self.actionPrevCol = QtGui.QAction('Prev column', self)
         self.actionPrevCol.setShortcut(Qt.SHIFT + Qt.Key_Left)
-        connect(self.actionPrevCol, SIGNAL('triggered()'),
-                self.textview_status.prevCol)
+        self.actionPrevCol.triggered.connect(self.textview_status.prevCol)
         self.addAction(self.actionPrevCol)
 
         # Activate file (file diff navigator)
@@ -319,13 +388,12 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
             if not isinstance(w, QtGui.QLineEdit):
                 self.tableView_filelist.fileActivated(self.tableView_filelist.currentIndex(),)
             else:
-                w.emit(SIGNAL('editingFinished()'))
-        connect(self.actionActivateFile, SIGNAL('triggered()'),
-                enterkeypressed)
+                w.editingFinished.emit()
+                self.actionActivateFile.triggered.connect(enterkeypressed)
 
         self.actionActivateFileAlt = QtGui.QAction('Activate alt. file', self)
         self.actionActivateFileAlt.setShortcuts([Qt.ALT+Qt.Key_Return, Qt.ALT+Qt.Key_Enter])
-        connect(self.actionActivateFileAlt, SIGNAL('triggered()'),
+        self.actionActivateFileAlt.triggered.connect(
                 lambda self=self:
                 self.tableView_filelist.fileActivated(self.tableView_filelist.currentIndex(),
                                                       alternate=True))
@@ -350,9 +418,8 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
 
     def create_models(self, fromhead=None):
         self.repomodel = HgRepoListModel(self.repo, fromhead=fromhead)
-        connect(self.repomodel, SIGNAL('filled'),
-                self.on_filled)
-        connect(self.repomodel, SIGNAL('showMessage'),
+        self.repomodel.filled.connect(self.on_filled)
+        self.repomodel.message_logged.connect(
                 self.statusBar().showMessage,
                 Qt.QueuedConnection)
 
@@ -378,13 +445,19 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
     def setupRevisionTable(self):
         view = self.tableView_revisions
         view.installEventFilter(self)
-        connect(view, SIGNAL('clicked (const QModelIndex &)'),
-                self.toggleMainContent)
-        connect(view, SIGNAL('revisionSelected'), self.revision_selected)
-        connect(view, SIGNAL('revisionActivated'), self.revision_activated)
-        connect(view, SIGNAL('startFromRev'), self.start_from_rev)
-        connect(self.textview_header, SIGNAL('revisionSelected'), view.goto)
-        connect(self.textview_header, SIGNAL('parentRevisionSelected'),
+        view.clicked.connect(self.toggleMainContent)
+        view.revision_selected.connect(self.revision_selected)
+        view.revision_selected[int].connect(self.revision_selected)
+        view.revision_activated.connect(self.revision_activated)
+        view.revision_activated[int].connect(self.revision_activated)
+        view.start_from_rev.connect(self.start_from_rev)
+        view.start_from_rev[int, bool].connect(self.start_from_rev)
+        view.start_from_rev[str, bool].connect(self.start_from_rev)
+        self.textview_header.revision_selected.connect(view.goto)
+        self.textview_header.revision_selected[int].connect(view.goto)
+        self.textview_header.parent_revision_selected.connect(
+                self.textview_status.displayDiff)
+        self.textview_header.parent_revision_selected[int].connect(
                 self.textview_status.displayDiff)
         self.attachQuickBar(view.goto_toolbar)
         gotoaction = view.goto_toolbar.toggleViewAction()
@@ -392,14 +465,9 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self.toolBar_edit.addAction(gotoaction)
 
     def start_from_rev(self, rev=None, follow=False):
-        if rev == self.startrev_entry:
-            rev = str(self.startrev_entry.text()).strip() or None
-        if follow == self.startrev_follow_action:
-            follow = self.startrev_follow_action.isChecked()
-
+        rev = rev or None # '' => None
         self.startrev_entry.setText(str(rev or ''))
         self.startrev_follow_action.setChecked(follow)
-
         self.refreshRevisionTable(rev=rev, follow=follow)
 
     def _setup_table(self, table):
@@ -436,7 +504,11 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         idx = tv.model().index(0, 0) # Working directory or tip
         if not wc.dirty() and wc.p1().rev() >= 0:
             # parent of working directory is not nullrev
-            idx = tv.model().indexFromRev(wc.p1().rev())
+            _idx = tv.model().indexFromRev(wc.p1().rev())
+            # may appears if wc has been filtered out
+            # (for example not on the filtered branch)
+            if _idx is not None:
+                idx = _idx
         tv.setCurrentIndex(idx)
 
     def revision_activated(self, rev=None):
@@ -449,7 +521,7 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self._manifestdlg = ManifestViewer(self.repo, rev)
         self._manifestdlg.show()
 
-    def revision_selected(self, rev):
+    def revision_selected(self, rev=None):
         """
         Callback called when a revision is selected in the revisions table
         if rev == -1: refresh the current selected revision
@@ -507,24 +579,24 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
             self.close()
 
     def ask_repository(self):
-        qrepopath = QtGui.QFileDialog.getExistingDirectory(
+        repopath = QtGui.QFileDialog.getExistingDirectory(
             self,
             'Select a mercurial repository',
             self.repo.root or os.path.expanduser('~'))
-        repopath = find_repository(str(qrepopath))
-        if not repopath:
+        repopath = find_repository(repopath)
+        if not (repopath or self.repo.root):
             if not self.repo.root:
                 raise RepoError("There is no Mercurial repository here (.hg not found)!")
             else:
-                return
-        repo = build_repo(ui.ui(), repopath)
-        return repo
+                return None
+        return repopath
 
-    def open_repository(self):
-        repo = self.ask_repository()
-        if not repo:
+    def open_repository(self, repopath=None):
+        if repopath is None:
+            repopath = self.ask_repository()
+        if repopath is None:
             return
-        self.repo = repo
+        self.repo = build_repo(ui.ui(), repopath)
         self.cfg = HgConfig(self.repo.ui)
         self.setWindowTitle('hgview: %s' % os.path.abspath(self.repo.root))
         self._finish_load()
@@ -540,11 +612,12 @@ class HgRepoViewer(QtGui.QMainWindow, HgDialogMixin, _HgRepoViewer):
         self._repodate = self._getrepomtime()
         self.setupBranchCombo()
         self.setupModels()
+        self._setupQuickOpen()
 
     #@timeit
     def refreshRevisionTable(self, *args, **kw):
         """Starts the process of filling the HgModel"""
-        branch = str(self.branch_comboBox.currentText())
+        branch = self.branch_comboBox.currentText()
         startrev = kw.get('rev', None)
         # XXX workaround: self.sender() may provoke a core dump if
         # this method is called directly (not via a connected signal);
